# MIT License
#
# Copyright (c) 2015-2022 Iakiv Kramarenko
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
import warnings
from typing import Any, Union, Iterable

from selene.core import match
from selene.core.condition import Condition
from selene.core.entity import Element, Collection, Browser
from selene.support.conditions import not_ as _not_

no = _not_


def exact_text(value) -> Condition[Element]:
    return match.element_has_exact_text(value)


# todo: consider accepting int
def text(partial_value) -> Condition[Element]:
    return match.element_has_text(partial_value)


# todo: should we use here js.property style (and below for js.returned(...))
def js_property(name: str, value: str = None):
    if value:
        warnings.warn(
            'passing second argument is deprecated; use have.js_property(foo).value(bar) instead',
            DeprecationWarning,
        )
        return match.element_has_js_property(name).value(value)

    return match.element_has_js_property(name)


def css_property(name: str, value: str = None):
    if value:
        warnings.warn(
            'passing second argument is deprecated; use have.css_property(foo).value(bar) instead',
            DeprecationWarning,
        )
        return match.element_has_css_property(name).value(value)

    return match.element_has_css_property(name)


def attribute(name: str, value: str = None):
    if value:
        warnings.warn(
            'passing second argument is deprecated; use have.attribute(foo).value(bar) instead',
            DeprecationWarning,
        )
        return match.element_has_attribute(name).value(value)

    return match.element_has_attribute(name)


def value(text) -> Condition[Element]:
    return match.element_has_value(text)


def values(*texts: Union[str, Iterable[str]]) -> Condition[Collection]:
    return match.collection_has_values(*texts)


def value_containing(partial_text) -> Condition[Element]:
    return match.element_has_value_containing(partial_text)


def values_containing(
    *partial_texts: Union[str, Iterable[str]]
) -> Condition[Collection]:
    return match.collection_has_values_containing(*partial_texts)


def css_class(name) -> Condition[Element]:
    return match.element_has_css_class(name)


def tag(name: str) -> Condition[Element]:
    return match.element_has_tag(name)


def tag_containing(name: str) -> Condition[Element]:
    return match.element_has_tag_containing(name)


# *** SeleneCollection conditions ***


def size(number: int) -> Condition[Collection]:
    return match.collection_has_size(number)


def size_less_than(number: int) -> Condition[Collection]:
    return match.collection_has_size_less_than(number)


def size_less_than_or_equal(number: int) -> Condition[Collection]:
    return match.collection_has_size_less_than_or_equal(number)


def size_greater_than(number: int) -> Condition[Collection]:
    return match.collection_has_size_greater_than(number)


def size_at_least(number: int) -> Condition[Collection]:
    warnings.warn(
        'might be deprecated; use have.size_greater_than_or_equal instead',
        PendingDeprecationWarning,
    )
    return match.collection_has_size_greater_than_or_equal(number)


def size_greater_than_or_equal(number: int) -> Condition[Collection]:
    return match.collection_has_size_greater_than_or_equal(number)


# todo: consider accepting ints
def texts(*partial_values: Union[str, Iterable[str]]) -> Condition[Collection]:
    return match.collection_has_texts(*partial_values)


def exact_texts(*values: Union[str, Iterable[str]]) -> Condition[Collection]:
    return match.collection_has_exact_texts(*values)


def url(exact_value: str) -> Condition[Browser]:
    return match.browser_has_url(exact_value)


def url_containing(partial_value: str) -> Condition[Browser]:
    return match.browser_has_url_containing(partial_value)


def title(exact_value: str) -> Condition[Browser]:
    return match.browser_has_title(exact_value)


def title_containing(partial_value: str) -> Condition[Browser]:
    return match.browser_has_title_containing(partial_value)


def tabs_number(value: int) -> Condition[Browser]:
    return match.browser_has_tabs_number(value)


def tabs_number_less_than(value: int) -> Condition[Browser]:
    return match.browser_has_tabs_number_less_than(value)


def tabs_number_less_than_or_equal(value: int) -> Condition[Browser]:
    return match.browser_has_tabs_number_less_than_or_equal(value)


def tabs_number_greater_than(value: int) -> Condition[Browser]:
    return match.browser_has_tabs_number_greater_than(value)


def tabs_number_greater_than_or_equal(value: int) -> Condition[Browser]:
    return match.browser_has_tabs_number_greater_than_or_equal(value)


def js_returned_true(script_to_return_bool: str) -> Condition[Browser]:
    warnings.warn(
        'might be deprecated; use have.js_returned(True, ...) instead',
        PendingDeprecationWarning,
    )
    return match.browser_has_js_returned(True, script_to_return_bool)


def js_returned(expected: Any, script: str, *args) -> Condition[Browser]:
    return match.browser_has_js_returned(expected, script, *args)



# MIT License
#
# Copyright (c) 2015-2022 Iakiv Kramarenko
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from __future__ import annotations

import atexit
import dataclasses
import itertools
import os
import time
from dataclasses import dataclass, field
from typing import Callable, Optional, Union

from selene.common import fp
from selene.common.fp import F

from selene.core.exceptions import TimeoutException
from selenium.webdriver.remote.webdriver import WebDriver

from selene.core.wait import Wait, E


@dataclass
class Config:
    """
    A one cross-cutting-concern-like object to group all options
    that might influence Selene behavior depending on context.
    As option the driver instance is also considered.
    More over, this config is not just config,
    but partially manages the driver lifecycle.

    All this makes it far from being a simple options data class...
    – yet kept as one «class for everything» to keep things easier to use,
    especially taking into account some historical reasons of Selene's design.
    """

    driver: Union[WebDriver, Callable[[], WebDriver]] = ...
    """
    A driver instance or a function that returns a driver instance.
    This driver will be quit on exit by default.
    To keep its finalization on your side, set ``self.hold_browser_open=True``.
    """

    @property
    def _driver_instance(self) -> WebDriver:
        """
        Depending on whether ``self.driver`` is instance or instance factory (function),
        returns driver instance correspondingly.
        """
        return self.driver() if callable(self.driver) else self.driver

    hold_browser_open: bool = False
    """Controls whether driver will be automatically quit on process exit or not."""

    def __post_init__(self):
        self._register_browser_finalizer()

    def _register_browser_finalizer(self):
        atexit.register(
            lambda: (
                self._driver_instance.quit()
                if not self.hold_browser_open and self._is_browser_alive
                else None
            )
        )

    @property
    def _is_browser_alive(self) -> bool:
        if self.driver is ...:
            return False

        try:
            return self._driver_instance.title is not None
        except Exception:  # noqa
            return False

    # TODO: consider accepting also hub url as "browser"
    #       because in case of "remote" mode, we will not need the common "name" like
    #       chrome or ff
    #       but we would pass the same name somewhere in caps... to choose correct "platform"
    #       so... then browser_name is kind of incorrect name becomes...
    #       why then not rename browser_name here to just browser...
    #       then... technically it might be possible to write something like:
    #           browser.config.browser = ... :)
    #              how can we make it impossible?
    #              or what else better name can we choose?
    #       ...
    #       if we are going to accept config.options
    #       where best it would be to put remote_url?
    #       config.service? config.driver? config.remote_url? config.executor?
    browser_name: str = 'chrome'
    """
    Desired name of the browser
    """

    timeout: float = 4
    base_url: str = ''

    set_value_by_js: bool = False
    type_by_js: bool = False
    click_by_js: bool = False
    wait_for_no_overlap_found_by_js: bool = False
    log_outer_html_on_failure: bool = False

    window_width: Optional[int] = None
    window_height: Optional[int] = None

    # TODO: better name? now technically it's not a decorator but decorator_builder...
    # or decorator_factory...
    # yet in python they call it just «decorator with args» or «decorator with params»
    # so technically I am correct naming it simply _wait_decorator
    # by type hint end users yet will see the real signature
    # and hence guess its «builder-like» nature
    # yet... should we for verbosity distinguish options
    # that a decorator factories from options that are simple decorators?
    # maybe better time to decide on this will be once we have more such options :p
    _wait_decorator: Callable[
        [Wait[E]], Callable[[F], F]
    ] = lambda _: fp.identity

    hook_wait_failure: Optional[Callable[[TimeoutException], Exception]] = None
    '''
    A handler for all exceptions, thrown on failed waiting for timeout.
    Should process the original exception and rethrow it or the modified one.

    TODO: why we name it as hook_* why not handle_* ?
          what would be proper style?
    '''

    poll_during_waits: int = 100
    """
    a fake option, not currently used in Selene waiting:)
    """
    reports_folder: str = field(
        default_factory=lambda: os.path.join(
            os.path.expanduser('~'),
            '.selene',
            'screenshots',
            str(round(time.time() * 1000)),
        ),
    )
    save_screenshot_on_failure: bool = True
    save_page_source_on_failure: bool = True
    _counter: itertools.count = field(
        default_factory=lambda: itertools.count(
            start=int(round(time.time() * 1000))
        )
    )
    """
    A counter, currently used for incrementing screenshot names
    """

    def with_(self, **config_as_kwargs) -> Config:
        return dataclasses.replace(self, **config_as_kwargs)

    def _generate_filename(self, prefix='', suffix=''):
        path = self.reports_folder
        next_id = next(self._counter)
        filename = f'{prefix}{next_id}{suffix}'
        file = os.path.join(path, f'{filename}')

        folder = os.path.dirname(file)
        if not os.path.exists(folder) and folder:
            os.makedirs(folder)

        return file

    # TODO: consider moving this injection to the WaitingEntity.wait method to build Wait object instead of config.wait
    def _inject_screenshot_and_page_source_pre_hooks(self, hook):
        # todo: consider moving hooks to class methods accepting config as argument
        #       or refactor somehow to eliminate all times defining hook fns
        def save_and_log_screenshot(error: TimeoutException) -> Exception:
            from selene.support.webdriver import WebHelper

            path = WebHelper(self.driver).save_screenshot(
                self._generate_filename(suffix='.png')
            )
            self.last_screenshot = path
            return TimeoutException(
                error.msg
                + f'''
Screenshot: file://{path}'''
            )

        def save_and_log_page_source(error: TimeoutException) -> Exception:
            filename = (
                # TODO: this dependency to last_screenshot might lead to code,
                #       when wrong screenshot name is taken
                self.last_screenshot.replace('.png', '.html')
                if getattr(self, 'last_screenshot', None)
                else self._generate_filename(suffix='.html')
            )
            from selene.support.webdriver import WebHelper

            path = WebHelper(self.driver).save_page_source(filename)
            self.last_page_source = path
            return TimeoutException(error.msg + f'\nPageSource: file://{path}')

        return fp.pipe(
            save_and_log_screenshot
            if self.save_screenshot_on_failure
            else None,
            save_and_log_page_source
            if self.save_page_source_on_failure
            else None,
            hook,
        )

    # TODO: we definitely not need it inside something called Config, especially "base interface like config
    def wait(self, entity):
        hook = self._inject_screenshot_and_page_source_pre_hooks(
            self.hook_wait_failure
        )
        return Wait(
            entity,
            at_most=self.timeout,
            or_fail_with=hook,
            _decorator=self._wait_decorator,
        )

    # def as_dict(self):
    #     return dict(
    #         (field.name, getattr(self, field.name))
    #         for field in dataclasses.fields(self)  # noqa
    #     )



# MIT License
#
# Copyright (c) 2015-2022 Iakiv Kramarenko
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
from typing import List, Dict, Any, Union

from selene.core.entity import Browser, Element, Collection
from selene.core.wait import Query


def attribute(name: str) -> Query[Element, str]:
    def fn(element: Element) -> str:
        return element().get_attribute(name)

    return Query(f'attribute {name}', fn)


inner_html = attribute('innerHTML')


outer_html = attribute('outerHTML')


text_content = attribute('textContent')
"""
full text of element without space normalization
"""


value = attribute('value')

tag: Query[Element, str] = Query(
    'tag name', lambda element: element().tag_name
)

text: Query[Element, str] = Query('text', lambda element: element().text)
"""
normalized text of element
"""

# todo: do we need condition for the following?
location_once_scrolled_into_view: Query[Element, Dict[str, int]] = Query(
    'location once scrolled into view',
    lambda element: element().location_once_scrolled_into_view,
)

# todo: what to do now with have.size* ? o_O
size: Union[Query[Element, Dict[str, Any]], Query[Collection, int]] = Query(
    'size',
    lambda entity: entity().size
    if isinstance(entity, Element)
    else len(entity()),
)

# todo: do we need condition for the following?
location: Query[Element, Dict[str, int]] = Query(
    'location', lambda element: element().location
)

# todo: do we need condition for the following?
rect: Query[Element, Dict[str, Any]] = Query(
    'rect', lambda element: element().rect
)

screenshot_as_base64: Query[Element, Any] = Query(
    'screenshot as base64', lambda element: element().screenshot_as_base64
)

screenshot_as_png: Query[Element, Any] = Query(
    'screenshot as png', lambda element: element().screenshot_as_png
)


def screenshot(filename: str) -> Query[Element, bool]:
    def fn(element: Element) -> str:
        return element().screenshot(filename)

    return Query(f'screenshot {filename}', fn)


# not needed, because interfere with "parent element" meaning and usually can be workaround via element.config.driver
# parent: Query[Element, Any] = \
#     Query('parent', lambda element: element().parent)
# todo: but should we add it with another name?


internal_id: Query[Element, Any] = Query(
    'internal id', lambda element: element().id
)


def css_property(name: str) -> Query[Element, str]:
    def fn(element: Element) -> str:
        return element().value_of_css_property(name)

    return Query(f'css property {name}', fn)


def js_property(name: str) -> Query[Element, str]:
    def fn(element: Element) -> str:
        return element().get_property(name)

    return Query(f'js property {name}', fn)


# --- Collection queries --- #

# --- Browser queries --- #


url: Query[Browser, str] = Query(
    'url', lambda browser: browser.driver.current_url
)

title: Query[Browser, str] = Query(
    'title', lambda browser: browser.driver.title
)

tabs: Query[Browser, List[str]] = Query(
    'tabs', lambda browser: browser.driver.window_handles
)

tabs_number: Query[Browser, int] = Query(
    'tabs number', lambda browser: len(browser.driver.window_handles)
)


def tab(index: int) -> Query[Browser, str]:
    def fn(browser: Browser) -> str:
        return browser.driver.window_handles[index]

    return Query(f'tab by index {index}', fn)


current_tab: Query[Browser, str] = Query(
    'current tab (window handle)',
    lambda browser: browser.driver.current_window_handle,
)


def __next_tab_fn(browser: Browser) -> str:
    tabs = browser.driver.window_handles
    current = browser.driver.current_window_handle
    current_index = tabs.index(current)
    return (
        tabs[0] if current_index >= len(tabs) - 1 else tabs[current_index + 1]
    )


next_tab: Query[Browser, str] = Query('next tab', __next_tab_fn)


def __previous_tab_fn(browser: Browser) -> str:
    tabs = browser.driver.window_handles
    current = browser.driver.current_window_handle
    current_index = tabs.index(current)
    return tabs[-1] if current_index == 0 else tabs[current_index - 1]


previous_tab: Query[Browser, str] = Query('previous tab', __previous_tab_fn)


page_source: Query[Browser, str] = Query(
    'page source', lambda browser: browser.driver.page_source
)



# MIT License
#
# Copyright (c) 2015-2022 Iakiv Kramarenko
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
import warnings
from typing import Union, Tuple, List, Type, Iterable, Any

from dataclasses import dataclass
from selenium.webdriver.common.by import By


def as_dict(o, skip_empty=True):
    return (
        {
            k: v
            for k, v in o.__dict__.items()
            if not (skip_empty and v is None) and not k.startswith('_')
        }
        if o
        else {}
    )


def to_by(selector_or_by: Union[str, tuple]) -> Tuple[str, str]:
    # todo: will it work `if isinstance(css_selector_or_by, Tuple[str, str]):` ?
    if isinstance(selector_or_by, tuple):
        return selector_or_by
    if isinstance(selector_or_by, str):
        return (
            (By.XPATH, selector_or_by)
            if (
                selector_or_by.startswith('/')
                or selector_or_by.startswith('./')
                or selector_or_by.startswith('..')
                or selector_or_by.startswith('(')
            )
            else (By.CSS_SELECTOR, selector_or_by)
        )
    raise TypeError(
        'selector_or_by should be str with CSS selector or XPATH selector or Tuple[by:str, value:str]'
    )


def flatten(collection: Iterable[Union[Iterable[Any], Any]]) -> Iterable[Any]:
    # todo: consider adding skip_none=False option
    return tuple(  # TODO: refactor to return tuple
        item_or_inner
        for item in collection
        for item_or_inner in (
            item
            if (isinstance(item, Iterable) and not isinstance(item, str))
            else [item]
        )
    )


def dissoc(associated: dict, *keys: str) -> dict:
    return {k: v for k, v in associated.items() if k not in keys}


def on_error_return_false(no_args_predicate):
    try:
        return no_args_predicate()
    except Exception:
        return False


def is_absolute_url(relative_or_absolute_url: str) -> bool:
    url = relative_or_absolute_url.lower()
    return (
        url.startswith('http:')
        or url.startswith('https:')
        or url.startswith('file:')
        or url.startswith('about:')
        or url.startswith('data:')
    )



from __future__ import annotations

import atexit
import re
import warnings
from dataclasses import field
from typing import Callable, Union, Optional

from selenium.webdriver.remote.webdriver import WebDriver

from selene.core.configuration import Config as BaseConfig
from selene.core.entity import Browser


class Config(BaseConfig):
    _driver: WebDriver = field(default=..., init=False, repr=False)
    _driver_source: Callable[[], WebDriver] = field(
        default=..., init=False, repr=False
    )

    @property
    def driver(self) -> WebDriver:
        self._driver = self._driver_source()
        return self._driver

    @driver.setter
    def driver(self, value: Union[WebDriver, Callable[[], WebDriver]]):
        def ensure_installed_and_built_when_not_alive():
            return (
                self._driver
                if (self._driver is not ... and self._is_browser_alive)
                else _install_and_build_driver(self.browser_name)
            )

        if (
            # driver was previously set via built-in driver-management logic
            re.sub(' at 0x.+>$', '>', str(self._driver_source))
            == re.sub(
                ' at 0x.+>$',
                '>',
                str(ensure_installed_and_built_when_not_alive),
            )
            # and ...
            and self._driver is not ...
            and self._is_browser_alive
        ):
            self._driver.quit()

        if value is ... or value is None:
            self._driver = ...
            self._driver_source = ensure_installed_and_built_when_not_alive
        elif callable(value):
            self._driver = ...
            self._driver_source = value
        else:
            self._driver = value
            self._driver_source = lambda: value

        atexit.register(
            lambda: (
                self.driver.quit()
                if not self.hold_browser_open
                and self._driver is not ...
                and self._is_browser_alive
                else None
            )
        )

    def get_or_create_driver(self) -> WebDriver:
        warnings.warn(
            'config.get_or_create_driver is deprecated, use config.driver instead',
            DeprecationWarning,
        )
        return self.driver

    def _reset_driver_source(self):
        """
        set default value (...) to the self.driver and so to the self._driver
        Hence, by this – also resetting self._driver_source,
        i.e. returning its original «automatically managed by Selene» version

        TODO: should we make this method public?
              currently the official way to reset driver source for the user
              is `browser.config.driver = ...`
              wouldn't it be better to allow him to do:
              is `browser.config.reset_driver_source()`
              ?
        """
        self.driver = ...

    def reset_driver(self):  # noqa
        warnings.warn(
            'selene.browser.config.reset_driver is deprecated '
            'in favor of `config.driver = ...`',
            DeprecationWarning,
        )
        self.driver = ...

    @property
    def _is_browser_alive(self) -> bool:
        if self._driver is ...:
            # raise _SeleneError('the driver has not been sourced yet at config')
            warnings.warn(
                'The driver has not been sourced yet at config, '
                'but asked for being alive. '
                'There might be a situation when driver is already created '
                'and so alive but passed to config via callable '
                'and not sourced yet at config...'
            )
            return False

        try:
            return self._driver.title is not None
        except Exception:  # noqa
            return False

    def __post_init__(self):
        # TODO: consider making private
        self.last_screenshot: Optional[str] = None
        self.last_page_source: Optional[str] = None


def _install_and_build_driver(browser_name):
    # todo: do we need here pass self.desired_capabilities too?

    from selenium.webdriver import ChromeOptions, Chrome, Firefox, Edge
    from selenium.webdriver.chrome.service import Service as ChromeService
    from selenium.webdriver.edge.service import Service as EdgeService
    from selenium.webdriver.firefox.service import Service as FirefoxService
    from webdriver_manager.chrome import ChromeDriverManager
    from webdriver_manager.firefox import GeckoDriverManager
    from webdriver_manager.microsoft import EdgeChromiumDriverManager
    from webdriver_manager.core.utils import ChromeType

    def get_chrome():
        return Chrome(
            service=ChromeService(
                ChromeDriverManager(chrome_type=ChromeType.GOOGLE).install()
            ),
            options=ChromeOptions(),
        )

    def get_firefox():
        return Firefox(service=FirefoxService(GeckoDriverManager().install()))

    def get_edge():
        return Edge(service=EdgeService(EdgeChromiumDriverManager().install()))

    return {
        'chrome': get_chrome,
        'firefox': get_firefox,
        'edge': get_edge,
    }.get(browser_name)()


config = Config()
browser = Browser(config)



# MIT License
#
# Copyright (c) 2015-2022 Iakiv Kramarenko
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
from typing import Union

from selene.core.entity import Element
from selene.core.wait import Command


# noinspection PyPep8Naming
class js:
    @staticmethod
    def set_value(value: Union[str, int]) -> Command[Element]:
        def fn(element: Element):
            element.execute_script(
                """
                var text = arguments[0];
                var maxlength = element.getAttribute('maxlength') === null
                    ? -1
                    : parseInt(element.getAttribute('maxlength'));
                element.value = maxlength === -1
                    ? text
                    : text.length <= maxlength
                        ? text
                        : text.substring(0, maxlength);
                return null;
                """,
                str(value),
            )

        return Command(f'set value by js: {value}', fn)

    @staticmethod
    def type(keys: Union[str, int]) -> Command[Element]:
        def fn(element: Element):
            element.execute_script(
                """
                textToAppend = arguments[0];
                var value = element.value || '';
                var text = value + textToAppend;
                var maxlength = element.getAttribute('maxlength') === null
                    ? -1
                    : parseInt(element.getAttribute('maxlength'));
                element.value = maxlength === -1
                    ? text
                    : text.length <= maxlength
                        ? text
                        : text.substring(0, maxlength);
                return null;
                """,
                str(keys),
            )

        return Command(f'set value by js: {keys}', fn)

    scroll_into_view = Command(
        'scroll into view',
        lambda element: element.execute_script('element.scrollIntoView(true)'),
    )

    click = Command(
        'click',
        lambda element: element.execute_script('element.click()'),
    )

    clear_local_storage = Command(
        'clear local storage',
        lambda browser: browser.driver.execute_script(
            'window.localStorage.clear()'
        ),
    )

    clear_session_storage = Command(
        'clear local storage',
        lambda browser: browser.driver.execute_script(
            'window.sessionStorage.clear()'
        ),
    )

    remove = Command(
        'remove',
        lambda entity: (
            entity.execute_script('element.remove()')
            if not hasattr(entity, '__iter__')
            else [
                element.execute_script('element.remove()')
                for element in entity
            ]
        ),
    )

    @staticmethod
    def set_style_property(
        name: str, value: Union[str, int]
    ) -> Command[Element]:
        return Command(
            f'set element.style.{name}="{value}"',
            lambda entity: (
                entity.execute_script(f'element.style.{name}="{value}"')
                if not hasattr(entity, '__iter__')
                else [
                    element.execute_script(f'element.style.{name}="{value}"')
                    for element in entity
                ]
            ),
        )

    set_style_display_to_none = Command(
        'set element.style.display="none"',
        lambda entity: (
            entity.execute_script('element.style.display="none"')
            if not hasattr(entity, '__iter__')
            else [
                element.execute_script('element.style.display="none"')
                for element in entity
            ]
        ),
    )

    set_style_display_to_block = Command(
        'set element.style.display="block"',
        lambda entity: (
            entity.execute_script('element.style.display="block"')
            if not hasattr(entity, '__iter__')
            else [
                element.execute_script('element.style.display="block"')
                for element in entity
            ]
        ),
    )

    set_style_visibility_to_hidden = Command(
        'set element.style.visibility="hidden"',
        lambda entity: (
            entity.execute_script('element.style.visibility="hidden"')
            if not hasattr(entity, '__iter__')
            else [
                element.execute_script('element.style.visibility="hidden"')
                for element in entity
            ]
        ),
    )

    set_style_visibility_to_visible = Command(
        'set element.style.visibility="visible"',
        lambda entity: (
            entity.execute_script('element.style.visibility="visible"')
            if not hasattr(entity, '__iter__')
            else [
                element.execute_script('element.style.visibility="visible"')
                for element in entity
            ]
        ),
    )
